/* linmalloc.c
 * Written by Fredrik Hubinette (hubbe@lysator.liu.se)
 * This malloc is still experimental, it was designed to minimize
 * unnecessary paging. Thus: no routines traverses any linked lists.
 * It may behave extremly ugly with lots of mallocs in the 1-8 bytes area
 * (when I say lots I mean LOTS but don't say I didn't warn you)
 *
 * Overhead: 1 pointer per block
 *
 * Future extensions:
 * It should allocate small blocks together.
 *
 */

/* #define MDEBUG */

#include <string.h>
#include <sys/types.h>


/* if you don't define one of this, a hybrid fit will be used */
#define FIRST_FIT

/*
#define BEST_FIT
*/


#define START_PTRS 10

/* how many free lists should we use? */
#define MAX_FREE_LIST 20000

/* when splitting a block the rest is smaller than WASTELIMIT*4 bytes
 * the block won't be splitted */
#define WASTELIMIT 2

/* how many steps should we search the free_list ? */
#define LISTSEARCH (WASTELIMIT+used_ptrs>>10)

/* why sbrk small areas? */
#define MIN_SBRK 8192

/* declarations */
struct free_block
{
  struct free_block *next,**prevptr;
};

typedef struct free_block *free_block_ptr;
typedef void *memory;
typedef unsigned int bptr;

memory realloc(memory,int);
memory malloc(int);
static int findpos(int);

/* handy-macros */
#define MEM_FREE 0x80000000
#define MASK     0x3fffffff
#define BPTR(X) (((int)(X))>>2)
#define APTR(X) ((memory)((X)<<2))
#define MIN(X,Y) ((X)<(Y)?(X):(Y))
#define MAX(X,Y) ((X)>(Y)?(X):(Y))

/* data */
int debugmalloc=0; /* not used  (yet) */
static int reallocing=0;
static int ptrs=0;
static int used_ptrs=0;
static bptr *mems=NULL;
static free_block_ptr free_blocks[MAX_FREE_LIST];

/* this one handles overlaps */
static void move(int *to,int *from,int ints)
{
  if(to<from)
  {
    while(ints--) *(to++)=*(from++);
  }else if(to>from){
    to+=ints-1;
    from+=ints-1;
    while(ints--) *(to--)=*(from--);
  }
}


static void crash_and_burn(char *s)
{
  write(2,s,strlen(s));
  abort();
}

#ifdef MDEBUG
static void check_mems_list()
{
  free_block_ptr p;
  int e,pos,tmp;
  if(!ptrs) return;
  for(e=0;e<used_ptrs-1;e++)
  {
    if( (mems[e]&MASK) >= (mems[e+1]&MASK) )
      crash_and_burn("Error in memory table.\n");
  }

  if(mems[used_ptrs-1] & MEM_FREE)
    crash_and_burn("Last pointer appears free.\n");

  for(e=0;e<MAX_FREE_LIST;e++)
  {
    for(p=free_blocks[e];p;p=p->next)
    {
      if(p->prevptr[0]!=p)
	crash_and_burn("Prevpointer doesn't point at a pointer to this block.\n");
      pos=findpos(BPTR(p));
      if(pos==-1)
	crash_and_burn("Error in free list.\n");
      if(!(mems[pos] & MEM_FREE))
	crash_and_burn("Memory in free list not free.\n");
      tmp=(mems[pos+1] & MASK) - (mems[pos] & MASK);
      if(tmp!=e)
	crash_and_burn("Wrong size of block in free list.\n");
    }
  }
}
#endif

static void unlink_free(int pos)
{
  int size;
  free_block_ptr f;
  size=(mems[pos+1] & MASK) - (mems[pos] & MASK);
  if(size<BPTR(sizeof(struct free_block)) || size>=MAX_FREE_LIST) return;
  f=(free_block_ptr)APTR(mems[pos] & MASK);
  *(f->prevptr)=f->next;
  if(f->next)
    f->next->prevptr=f->prevptr;
}

static void link_free(int pos)
{
  int size;
  free_block_ptr f;
  size=(mems[pos+1] & MASK) - (mems[pos] & MASK);
  if(size<BPTR(sizeof(struct free_block)) || size>=MAX_FREE_LIST) return;
  f=(free_block_ptr)APTR(mems[pos] & MASK);
#ifdef MDEBUG
  {
    free_block_ptr p;
    for(p=free_blocks[size];p;p=p->next)
      if(p==f)crash_and_burn("Linking block into free list again.\n");
  }
#endif
  if(free_blocks[size])
    free_blocks[size]->prevptr=&(f->next);
  f->next=free_blocks[size];
  free_blocks[size]=f;
  f->prevptr=free_blocks+size;
  
}

/* binary search : O(2log(used_ptrs)) */
static int findpos(int tmp)
{
  int a,b,c,d;
  a=0;
  b=used_ptrs;
  while(a!=b)
  {
    c=(a+b)>>1;
    d=mems[c] & MASK;
    if(d==tmp) return c;
    if(tmp<d) b=c; else a=c+1;
  }
  return -1;
}

static memory low_malloc(int foosize,int pos)
{
  int tmp;
  memory ret;
  tmp=(mems[pos+1] & MASK) - (mems[pos] & MASK);
  if(tmp<foosize) return NULL;
  unlink_free(pos);
  mems[pos]&=~MEM_FREE;
  if(tmp-foosize>WASTELIMIT)
  {
    /* we split the block here, take out the axe */
    move(mems+pos+2,mems+pos+1,used_ptrs-pos);
    used_ptrs++;
    mems[pos+1]=(mems[pos]+foosize) | MEM_FREE;
    link_free(pos+1);
  }
  ret=APTR(mems[pos]);
#ifdef MDEBUG
  check_mems_list();
#endif
  if(ptrs-used_ptrs<10 && !reallocing) /* I want my margins */
  {
    bptr *new,*old;

    ptrs*=2;
    reallocing=1;
    new=(bptr *)malloc(ptrs*sizeof(bptr));
    move(new,mems,used_ptrs);
    old=mems;
    mems=new;
    free(old);
    reallocing=0;
  }
#ifdef MDEBUG
  check_mems_list();
#endif
  return ret;
}


memory malloc(int size)
{
  int pos,foosize;
  memory new;

  foosize=BPTR(size+3);

  if(foosize<MAX_FREE_LIST && ptrs)
  {
    int tmp;
    free_block_ptr f;

    for(tmp=foosize;tmp<=MIN(foosize+LISTSEARCH,MAX_FREE_LIST);tmp++)
      if(f=free_blocks[tmp])
        if(-1!=(pos=findpos(BPTR(f))))
          if(new=low_malloc(foosize,pos))
            return new;
  }

#ifdef FIRST_FIT
    for(pos=0;pos<used_ptrs-1;pos++)
    {
      if(!(mems[pos] & MEM_FREE)) continue;
      if(new=low_malloc(foosize,pos)) return new;
    }
#else
#ifdef BEST_FIT
  {
    int bestpos,bestsize,tmp;
    bestsize=0x7fffffff;
    bestpos=0;
  
    for(pos=0;pos<used_ptrs-1 && bestsize!=foosize;pos++)
    {
      if(!(mems[pos] & MEM_FREE)) continue;
      tmp=(mems[pos+1] & MASK) - (mems[pos] & MASK);
      if(tmp<foosize) continue;
      if(tmp>=bestsize) continue;
      bestpos=pos;
      bestsize=tmp;
    }
    if(bestpos && (new=low_malloc(foosize,bestpos)))
       return new;
  }
#else
  {
    /* hybrid fit:
     * this is a best fit that just tries to avoid adding entries to
     * mems[], it does so by finding a free block like this:
     * size<=blocksize<=size+WASTELIMIT*4
     * Should work pretty fine if WASTELIMIT is tuned.
     */
    int bestpos,bestsize,tmp;
    bestsize=0x7fffffff;
    bestpos=0;
  
    for(pos=0;pos<used_ptrs-1 && bestsize>foosize+WASTELIMIT;pos++)
    {
      if(!(mems[pos] & MEM_FREE)) continue;
      tmp=(mems[pos+1] & MASK) - (mems[pos] & MASK);
      if(tmp<foosize) continue;
      if(tmp>=bestsize) continue;
      bestpos=pos;
      bestsize=tmp;
    }
    if(bestpos && (new=low_malloc(foosize,bestpos)))
       return new;
  }
#endif
#endif

  /* no memory found, pray to vmunix for more */
  if(!ptrs)
  {
    int high;
    ptrs=START_PTRS;

    mems=(bptr *)sbrk(sizeof(bptr)*ptrs+100);
    mems[0]=BPTR(mems);
    mems[1]=BPTR(mems+ptrs) | MEM_FREE;
    mems[2]=BPTR(sbrk(0));
    used_ptrs=3;
    link_free(1);
  }
  if(!(mems[used_ptrs-2] & MEM_FREE))
  {
    mems[used_ptrs]=mems[used_ptrs-1];
    mems[used_ptrs-1]|=MEM_FREE;
    used_ptrs++;
  }else{
    unlink_free(used_ptrs-2);
  }
  sbrk(MAX(size,MIN_SBRK));
  mems[used_ptrs-1]=BPTR(sbrk(0));
  link_free(used_ptrs-2);

  return low_malloc(foosize,used_ptrs-2);
}

static low_free(int pos)
{
  int size;
#ifdef MDEBUG
  check_mems_list();
#endif
  if(pos && (mems[pos-1] & MEM_FREE))
  {
    unlink_free(pos-1);
    if((pos+1)<used_ptrs && (mems[pos+1] & MEM_FREE))
    {
      unlink_free(pos+1);
      move(mems+pos,mems+pos+2,used_ptrs-pos-2);
      used_ptrs-=2;
    }else{
      move(mems+pos,mems+pos+1,used_ptrs-pos-1);
      used_ptrs--;
    }
    pos--;
  }else{
    mems[pos]|=MEM_FREE;
    if((pos+1)<used_ptrs && (mems[pos+1] & MEM_FREE))
    {
      unlink_free(pos+1);
      move(mems+pos+1,mems+pos+2,used_ptrs-pos-2);
      used_ptrs--;
    }
  }
#ifdef MDEBUG
  check_mems_list();
#endif
  link_free(pos);
#ifdef MDEBUG
  check_mems_list();
#endif
}

int free(memory p)
{
  int tmp,pos;
  tmp=BPTR(p);
  pos=findpos(tmp);

  if(pos==-1) crash_and_burn("Freeing nonexistant memory.\n");
  if(mems[pos] & MEM_FREE) crash_and_burn("Freeing already free memory.\n");
  low_free(pos);
  return 0;
}

memory realloc(memory p,int size)
{
  int tmp,pos,foosize,tmp2,tmp3;
  memory new;
#ifdef MDEBUG
  check_mems_list();
#endif

  tmp=BPTR(p);
  pos=findpos(tmp);

  if(pos==-1) crash_and_burn("reallocing nonexistant memory.\n");
  if(mems[pos] & MEM_FREE) crash_and_burn("realloc on free memory.\n");

  foosize=BPTR(size+3);

  tmp2=(mems[pos+1] & MASK) - tmp;

  if(mems[pos+1] & MEM_FREE)
  {
    tmp3=(mems[pos+2] & MASK) - tmp;
    if(tmp3>foosize)
    {
      unlink_free(pos+1);
      mems[pos+1]=(tmp+foosize) | MEM_FREE;
      link_free(pos+1);
      return p;
    }else if(tmp3==foosize){
      unlink_free(pos+1);
      move(mems+pos+1,mems+pos+2,used_ptrs-pos-2);
      used_ptrs--;
#ifdef MDEBUG
      check_mems_list();
#endif
      return p;
    }
  }else{
#if 1
    /* this may waste some memory */
    if(tmp2>=foosize) return p;
#else
    if(tmp2>=foosize && tmp2-foosize<WASTELIMIT) return p;
#endif
  }
  
#ifdef MDEBUG
  check_mems_list();
#endif
  new=malloc(size);
  move((int *)new,(int *)p,MIN(tmp2,foosize));
#ifdef MDEBUG
  check_mems_list();
#endif
  free(p);
#ifdef MDEBUG
  check_mems_list();
#endif
  return new;
}

memory calloc(int size,int nitems)
{
  memory new;
#ifdef MDEBUG
  check_mems_list();
#endif
  size*=nitems;
  new=malloc(size);
  if(new) memset(new,0,size);
#ifdef MDEBUG
  check_mems_list();
#endif
  return new;
}

void assert_malloced_memory(memory p)
{
  int pos;
  if(ptrs<used_ptrs)
    crash_and_burn("Not enough space for pointers!\n");
  pos=findpos(BPTR(p));
  if(pos==-1 || (mems[pos] & MEM_FREE) || pos>=used_ptrs)
    crash_and_burn("Assert malloced memory failed.\n");
}

dump_malloc_data() {}
